#include <ros/ros.h>
#include <can_msgs/Frame.h>
#include <nav_msgs/Odometry.h>
#include <std_msgs/Header.h>
#include <cstring>
#include <cstdint>
#include <vector>
#include <cmath>
#include <tf2/LinearMath/Quaternion.h>

ros::Publisher odom_pub;

// 定义缓存区，用来暂存底盘先发送的can帧，等待后续can帧都到达后一并进行转换并发送到ros话题。
std::vector<can_msgs::Frame> odom_buffer;

nav_msgs::Odometry can_frames_to_odometry(const std::vector<can_msgs::Frame> &frames)
{
    nav_msgs::Odometry odom;
    // 获取can帧数据
    std::vector<uint8_t> data_272, data_273, data_274;
    for (const auto &frame : frames)
    {
        std::vector<uint8_t> data;
        std::copy(frame.data.begin(), frame.data.end(), std::back_inserter(data));
        if (frame.id == 273)
        {
            data_273 = data;
        }
        else if (frame.id == 272)
        {
            data_272 = data;
        }
        else if (frame.id == 274)
        {
            data_274 = data;
        }
    }

    // 解析odom中X和Y的值
    float x, y, z;
    unsigned char temp[4];
    temp[0] = data_273[3];
    temp[1] = data_273[2];
    temp[2] = data_273[1];
    temp[3] = data_273[0];
    std::memcpy(&x, temp, sizeof(float));
    temp[0] = data_273[7];
    temp[1] = data_273[6];
    temp[2] = data_273[5];
    temp[3] = data_273[4];
    std::memcpy(&y, temp, sizeof(float));
    temp[0] = data_274[3];
    temp[1] = data_274[2];
    temp[2] = data_274[1];
    temp[3] = data_274[0];
    std::memcpy(&z, temp, sizeof(float));

    /*创建Odometry消息*/
    odom.header = std_msgs::Header();
    odom.header.seq = frames[0].header.seq;
    odom.header.stamp = frames[0].header.stamp;
    odom.header.frame_id = "odom";
    odom.child_frame_id = "base_link";
    /*给Odometry消息赋值*/
    odom.pose.pose.position.x = x / 1000; // 底盘在参考坐标系中的X坐标
    odom.pose.pose.position.y = y / 1000; // 底盘在参考坐标系中的Y坐标
    odom.pose.pose.position.z = 0.0;

    int16_t linear_vel_raw = (data_272[1] << 8) | data_272[2];
    int16_t angular_vel_raw = (data_272[3] << 8) | data_272[4];

    float linear_vel = static_cast<float>(linear_vel_raw) / 1000.0;     // 将can总线数据转换为实际值，单位为 m/s
    float angular_vel = static_cast<float>(angular_vel_raw) / 572.9578; // 将can总线数据转换为实际值，单位为 rad/s

    odom.twist.twist.linear.x = linear_vel; // 底盘线速度
    odom.twist.twist.linear.y = 0.0;
    odom.twist.twist.angular.z = angular_vel; // 底盘角速度

    // 机器人在参考坐标系中的朝向
    float orientation = z;

    // 将机器人的朝向角度限制在[-pi, pi)范围内
    // orientation = fmod(orientation + 180, 2.0 * 180) - 180;

    // 将机器人的朝向角度转换为四元数类型
    tf2::Quaternion q;
    q.setRPY(0, 0, orientation);

    // 将四元数类型赋值给ROS中的geometry_msgs::Quaternion类型的orientation变量
    geometry_msgs::Quaternion orientation_msg;
    odom.pose.pose.orientation.x = q.x();
    odom.pose.pose.orientation.y = q.y();
    odom.pose.pose.orientation.z = q.z();
    odom.pose.pose.orientation.w = q.w();

    return odom;
}

void received_messages_callback(const can_msgs::Frame::ConstPtr &msg)
{
    // 将接收到的can帧存储到缓存区
    odom_buffer.push_back(*msg);

    // 判断缓存区中是否同时收到了id为272、273和274的can帧
    bool found_272 = false, found_273 = false, found_274 = false;
    for (const auto &frame : odom_buffer)
    {
        if (frame.id == 272)
        {
            found_272 = true;
        }
        else if (frame.id == 273)
        {
            found_273 = true;
        }
        else if (frame.id == 274)
        {
            found_274 = true;
        }
    }

    // 如果同时收到了id为272、273和274的can帧，则转换成Odometry消息格式并发布到话题/chassis_odometry
    if (found_272 && found_273 && found_274)
    {
        nav_msgs::Odometry odom = can_frames_to_odometry(odom_buffer);

        std::cout << "The Odometry Data of small chassis: \n";
        std::cout << odom;

        odom_pub.publish(odom);

        // 清空缓存区
        odom_buffer.clear();
    }
}

int main(int argc, char **argv)
{
    ros::init(argc, argv, "can_to_odometry");
    ros::NodeHandle nh;

    ros::Subscriber received_messages_sub = nh.subscribe<can_msgs::Frame>("/received_messages", 10, received_messages_callback);
    odom_pub = nh.advertise<nav_msgs::Odometry>("/chassis_odometry", 10);

    ros::spin();
    return 0;
}