#include <ros/ros.h>
#include <can_msgs/Frame.h>
#include <socketcan_bridge/ViperCanData.h>
#include <std_msgs/String.h>
#include <iostream>

socketcan_bridge::ViperCanData parse_can_data_frame(const std::vector<uint8_t> &data_frame)
{
    if (data_frame.size() != 8)
        throw std::invalid_argument("Invalid data frame length"); // 如果收到的can数据帧的长度不等于8就报异常。

    socketcan_bridge::ViperCanData can_data;
    uint8_t motion_mode = data_frame[0];

    switch (motion_mode)
    {
    case 0x01:
        can_data.motion_mode = "Parking";
        break;
    case 0x02:
        can_data.motion_mode = "Spin";
        break;
    case 0x03:
        can_data.motion_mode = "Diagonal";
        break;
    case 0x04:
        can_data.motion_mode = "Ackermann";
        break;
    case 0x05:
        can_data.motion_mode = "Disabled";
        break;
    default:
        throw std::invalid_argument("Invalid motion mode");
    }

    can_data.remaining_battery = data_frame[5];

    uint8_t state_of_charge_and_scram_button = data_frame[6];
    switch (state_of_charge_and_scram_button)
    {
    case 0x01:
        can_data.state_of_charge_and_scram_button = "Charge--Running";
        break;
    case 0x00:
        can_data.state_of_charge_and_scram_button = "Discharge--Running";
        break;
    case 0x11:
        can_data.state_of_charge_and_scram_button = "charge--Stop";
        break;
    case 0x10:
        can_data.state_of_charge_and_scram_button = "Discharge--Stop";
        break;
    default:
        can_data.state_of_charge_and_scram_button = "Discharge--Running";
    }

    return can_data;
}

void received_messages_callback(const can_msgs::Frame::ConstPtr &msg, ros::Publisher &pub)
{
    if (msg->id == 272) // 如果id不是272则直接返回 return;
    {
        std::vector<uint8_t> data_frame(msg->data.begin(), msg->data.end());        // 将接收到的数据转化为字节向量。
        socketcan_bridge::ViperCanData can_data = parse_can_data_frame(data_frame); // 将字节向量转化成自定义消息类型ViperCanData。

        std::cout << "The motion state parameters of small chassis: \n";
        std::cout << can_data;

        pub.publish(can_data);
    }
}

int main(int argc, char **argv)
{
    ros::init(argc, argv, "viper_state");
    ros::NodeHandle nh;

    ros::Publisher can_data_pub = nh.advertise<socketcan_bridge::ViperCanData>("/chassis_state", 10);
    ros::Subscriber received_messages_sub = nh.subscribe<can_msgs::Frame>("/received_messages", 1, boost::bind(received_messages_callback, _1, can_data_pub));

    ros::spin();
    return 0;
}