/*
 * Copyright (c) 2016, Ivor Wanders
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *   * Redistributions of source code must retain the above copyright notice,
 *     this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *   * Neither the name of the copyright holder nor the names of its
 *     contributors may be used to endorse or promote products derived from
 *     this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#include <socketcan_bridge/topic_to_socketcan.h>
#include <socketcan_interface/string.h>
#include <geometry_msgs/Twist.h>
#include <string>
#include <ros/ros.h>
#include "sub.cpp"

namespace socketcan_bridge
{
  TopicToSocketCAN::TopicToSocketCAN(ros::NodeHandle *nh, ros::NodeHandle *nh_param, can::DriverInterfaceSharedPtr driver)
  {
    driver_ = driver;
    can_topic_ = nh->subscribe<geometry_msgs::Twist>("cmd_vel", nh_param->param("cmd_vel_queue_size", 10), std::bind(&TopicToSocketCAN::send_can_message, this, std::placeholders::_1));
    /*订阅/cmd_vel话题*/
    setup();
  };

  void TopicToSocketCAN::setup()
  {
    state_listener_ = driver_->createStateListener(
        std::bind(&TopicToSocketCAN::stateCallback, this, std::placeholders::_1));
  };

  void TopicToSocketCAN::send_can_message(const geometry_msgs::Twist::ConstPtr &msg)
  {
    /*回调函数send_can_message的功能是将从/cmd_vel话题上获得的线速度和角速度信息转换位16进制数后填入8位CAN数据帧的第二、三位和第四、五位，其中第二、四位表示线速度和角速度的高八位，第三、五位表示低八位。*/
    int16_t data_x = msg->linear.x * 1000;
    uint8_t linear_high = static_cast<uint8_t>((data_x >> 8) & 0xFF);
    uint8_t linear_low = static_cast<uint8_t>(data_x & 0xFF);
    int16_t data_z = msg->angular.z * 572.9578;
    uint8_t angular_high = static_cast<uint8_t>((data_z >> 8) & 0xFF);
    uint8_t angular_low = static_cast<uint8_t>(data_z & 0xFF);

    ros::param::get("mode", mode);

    can_msgs::Frame can_frame;
    can_frame.id = 0x100;
    can_frame.is_extended = false;
    can_frame.is_rtr = false;
    can_frame.is_error = false;
    can_frame.dlc = 8;
    can_frame.data[0] = mode;
    can_frame.data[1] = linear_high;
    can_frame.data[2] = linear_low;
    can_frame.data[3] = angular_high;
    can_frame.data[4] = angular_low;
    can_frame.data[5] = 0x00;
    can_frame.data[6] = 0x00;
    can_frame.data[7] = 0x00;

    can::Frame f;

    convertMessageToSocketCAN(can_frame, f);
    driver_->send(f);
    /*向CAN总线发送数据帧*/
    if (!f.isValid())
    {
      /* 检查 id 和标志是否合适。
       * ROS_WARN("Refusing to send invalid frame: %s.", can::tostring(f, true).c_str());
       * can::tostring 不能用于 > 8 帧。它会导致崩溃
       * due to usage of boost::array for the data array. The should always work.
       */
      ROS_ERROR("Invalid frame from topic: id: %#04x, length: %d, is_extended: %d", can_frame.id, can_frame.dlc, can_frame.is_extended);
      return;
    }

    bool res = driver_->send(f);
    if (!res)
    {
      ROS_ERROR("Failed to send message: %s.", can::tostring(f, true).c_str());
    }
  }

  void TopicToSocketCAN::stateCallback(const can::State &s)
  {
    std::string err;
    driver_->translateError(s.internal_error, err);
    if (!s.internal_error)
    {
      ROS_INFO("State: %s, asio: %s", err.c_str(), s.error_code.message().c_str());
    }
    else
    {
      ROS_ERROR("Error: %s, asio: %s", err.c_str(), s.error_code.message().c_str());
      setup();
    }
  };
}
